//------------------------------------------------------------------
// PS2 VU/VIF related functionality
// Low-level VIF/VU functions
//
// Geoff Audy, November 8th 2001
// Geoff Audy, July 5th 2002     Modified it to work under PS2Linux,
//
// Copyright (C)2002 Sony Computer Entertainment America Inc.
//
// This file is subject to the terms and conditions of the GNU
// General Public Licence Version 2.  See the file "COPYING"
// in the main directory of this archive for more details.
//------------------------------------------------------------------

#include <asm/types.h>           // Standard C Headers
#include <stdlib.h>              //
#include <ps2gs.h>               //
#include <ps2vpu.h>              //
#include <ps2vpufile.h>          //

#include "os.h"                  // Project Headers
#include "debug.h"               //
#include "ps2_dma.h"             //
#include "ps2_vif.h"             //

//-------------------------------------
#define  VIF_MINIMUMBUFFERSIZE  1024

typedef struct
{
   unsigned int  immediate:16;
   unsigned int  num:8;
   unsigned int  cmd:8;
} VIF_TAG  __attribute__ ((aligned (16)));;

//-------------------------------------
static unsigned int    VIF_Initialized = 0;     // Specifies if the module is initialized (1) or not (0)

static OS_Qword        *VIF_DoubleBuffer[2];    // The DMA chain (double-buffered)
static unsigned int    VIF_DoubleBufferSize[2]; // The DMA chain size (as a whole)

//----------------------------------------------------------
// Initialize the module.  A double-buffer must be provided,
// and must be big enough for its intended use, i.e. this
// module won't make any attempt at allocating more memory
// if the buffer is full.  Error messages will, however, be
// returned if this happens.  'doublebuffersize' is in bytes
// and 'doublebuffer' must be 16-byte aligned.
//----------------------------------------------------------
unsigned int VIF_ModuleInit (OS_Qword *doublebuffervif0,
                             unsigned int doublebuffervif0size,
                             OS_Qword *doublebuffervif1,
                             unsigned int doublebuffervif1size)
{
   // Don't initialize a module unless it's currently uninitialized
   if (VIF_Initialized == 1)
      return (VIF_RET_ALREADYINITIALIZED);

   // Make sure parameters are valid
   if ((doublebuffervif0 == NULL) ||
       (doublebuffervif0size < VIF_MINIMUMBUFFERSIZE) ||
       (doublebuffervif1 == NULL) ||
       (doublebuffervif1size < VIF_MINIMUMBUFFERSIZE))
      return (VIF_RET_INVALIDPARAMETER);

   // Initialize some local variables
   VIF_DoubleBuffer[0] = doublebuffervif0;
   VIF_DoubleBufferSize[0] = doublebuffervif0size;
   VIF_DoubleBuffer[1] = doublebuffervif1;
   VIF_DoubleBufferSize[1] = doublebuffervif1size;

   // Initialize the DMA chain for VIF0.  We're using the
   // DMA module for low-level DMA-chain management
   switch (DMA_ChainInit (VIF_DoubleBuffer[0], VIF_DoubleBufferSize[0],
                          DMA_CHANNEL_VIF0))
   {
      case DMA_RET_INVALIDPARAMETER:
            return (VIF_RET_INVALIDPARAMETER);

      case DMA_RET_WRONGALIGNMENT:
            return (VIF_RET_WRONGALIGNMENT);

      default:
            break;
   }

   // Initialize the DMA chain for VUF1.  We're using the
   // DMA module for low-level DMA-chain management
   switch (DMA_ChainInit (VIF_DoubleBuffer[1], VIF_DoubleBufferSize[1],
                          DMA_CHANNEL_VIF1))
   {
      case DMA_RET_INVALIDPARAMETER:
            return (VIF_RET_INVALIDPARAMETER);

      case DMA_RET_WRONGALIGNMENT:
            return (VIF_RET_WRONGALIGNMENT);

      default:
            break;
   }

#if 0
   // Reset VU1
   {
      asm __volatile__("
            addi    $8, $0, 768
            ctc2    $8, $vi28
            ");

      asm __volatile__("
            sync.l
            sync.p
            ");

      *VIF1_FBRST = 9;

      // Enable VIF1 interrupts, ignore DMATag mismatch errors, and consider those errors as VIFcode nops
      *VIF1_ERR = 6;
   }
#endif

   // From now on consider the module initialized
   VIF_Initialized = 1;

   // Everything was fine
   return (VIF_RET_OK);
}

//-------------------------------------------------------------
// Uninitialize the module.  Must have been initialized before.
//-------------------------------------------------------------
unsigned int VIF_ModuleUninit (void)
{
   if (VIF_Initialized == 0)
      return (VIF_RET_NOTINITIALIZED);

   // From now on consider the module uninitialized
   VIF_Initialized = 0;

   // Everything was fine
   return (VIF_RET_OK);
}

//-------------------------------------------------------------------
// Add code to be sent to VU memory via the DMA chain.  The data must
// be 16-byte aligned, and its size must be a multiple of 16 bytes.
//-------------------------------------------------------------------
unsigned int VIF_AddCode (OS_Qword *code, unsigned int dest, unsigned int forceupload)
{
   OS_Qword         *realcode;
   OS_Qword         *dmatag;
   unsigned int     codesize;
   static OS_Qword  *previouscode = (OS_Qword *)0xFFFFFFFF;

   if (VIF_Initialized == 0)
      return (VIF_RET_NOTINITIALIZED);

   // Make sure parameters are valid
   if ((code == NULL) || (dest >= VIF_DEST_MAX) || (forceupload >= VIF_UPLOAD_MAX))
      return (VIF_RET_INVALIDPARAMETER);

   // Only upload the code if we were forced to, or if it's different from the last code uploaded
   // (as there's no point in uploading the same code twice in a row)
   if ((forceupload == VIF_UPLOAD_FORCE) || (previouscode != code))
   {
      // Since the code is assumed to be preceded
      // by a dmatag, we've gotta skip it
      dmatag = (OS_Qword *)code;
      realcode = (OS_Qword *)code;
      realcode++;

      codesize = dmatag->ui32[0] & 0xFFFF;

      if (codesize == 0)
         return (VIF_RET_INVALIDPARAMETER);

      switch (DMA_ChainAdd (VIF_DoubleBuffer[dest], realcode,
                            codesize, DMA_ATTACHMENT_INLINE/*DMA_ATTACHMENT_REFERENCE*/))    // If you can guarantee all VU codes are compiled in an ELF, reference is better
      {
         case DMA_RET_INVALIDPARAMETER:
               return (VIF_RET_INVALIDPARAMETER);
         case DMA_RET_WRONGALIGNMENT:
               return (VIF_RET_WRONGALIGNMENT);
         case DMA_RET_DATATOOBIGTOFIT:
               DEBUG_Print ("ERROR: VIF buffer overflow\n");
               return (VIF_RET_DATATOOBIGTOFIT);
         default:
               break;
      }

      previouscode = code;
   }

   // Everything was fine
   return (VIF_RET_OK);
}

//----------------------------------------------------------------------------
// Add data to be sent to the VIF via the DMA chain.  The data must be 16-byte
// aligned, and its size must be specified in multiple of 16 bytes.
//----------------------------------------------------------------------------
unsigned int VIF_AddData (OS_Qword *buffer, unsigned int buffersizeinqw,
                          unsigned int attachment, unsigned int dest)
{
   if (VIF_Initialized == 0)
      return (VIF_RET_NOTINITIALIZED);

   // Make sure parameters are valid
   if ((buffer == NULL) || (buffersizeinqw == 0) ||
       (attachment >= VIF_ATTACHMENT_MAX) || (dest >= VIF_DEST_MAX))
      return (VIF_RET_INVALIDPARAMETER);

   // Make sure the alignment and size are proper
   if ((unsigned int)buffer & 0x0000000F)
      return (VIF_RET_WRONGALIGNMENT);

   // Add the data to the DMA chain (NOTE: Here I assume VIF_ATTACHENT_xxx
   // are the same as DMA_ATTACHMENT_xxx.  Bad me)
   switch (DMA_ChainAdd (VIF_DoubleBuffer[dest], buffer,
                         buffersizeinqw, attachment))
   {
      case DMA_RET_INVALIDPARAMETER:
            return (VIF_RET_INVALIDPARAMETER);
      case DMA_RET_WRONGALIGNMENT:
            return (VIF_RET_WRONGALIGNMENT);
      case DMA_RET_DATATOOBIGTOFIT:
            DEBUG_Print ("ERROR: VIF buffer overflow\n");
            return (VIF_RET_DATATOOBIGTOFIT);
      default:
            break; 
   }

   // Everything was fine
   return (VIF_RET_OK);
}

//--------------------------------------------------------------------
// Add VU data to be sent to the VIF via the DMA chain.  The data must
// be 16-byte aligned, and its size must be a multiple of 16 bytes.
// Note: It is better to use 'VIF_AddData' and have a VIFTag in front
//       of your data
//--------------------------------------------------------------------
unsigned int VIF_AddVUData (OS_Qword *buffer, unsigned int buffersizeinqw,
                            unsigned int destvuaddr, unsigned int attachment,
                            unsigned int dest)
{
   static unsigned int  vumemsize[2] = {256, 1024};
   static VIF_TAG       viftag[4] __attribute__ ((aligned (16))) = {{0, 0, VIF_CMD_NOP},
                                                                    {0, 0, VIF_CMD_NOP},
                                                                    {0, 0, VIF_CMD_NOP},
                                                                    {0, 0, VIF_CMD_UNPACK | VIF_UNPACK_V4_32}};

   if (VIF_Initialized == 0)
      return (VIF_RET_NOTINITIALIZED);

   // Make sure parameters are valid
   if ((buffer == NULL) || (buffersizeinqw == 0) ||
       (attachment >= VIF_ATTACHMENT_MAX) || (dest >= VIF_DEST_MAX) ||
       ((destvuaddr + buffersizeinqw) > vumemsize[dest]))
      return (VIF_RET_INVALIDPARAMETER);

   // Make sure the alignment and size are proper
   if ((unsigned int)buffer & 0x0000000F)
      return (VIF_RET_WRONGALIGNMENT);

   while (buffersizeinqw)
   {
      if (buffersizeinqw > 255)
      {
         // We're trying to send too much data in one go...  Cut it in smaller chunks

         viftag[3].num = 255;
         viftag[3].immediate = destvuaddr;

         // Add the tag to the DMA chain
         switch (DMA_ChainAdd (VIF_DoubleBuffer[dest], (OS_Qword *)viftag,
                               sizeof (viftag) / sizeof (OS_Qword), DMA_ATTACHMENT_INLINE))
         {
            case DMA_RET_INVALIDPARAMETER:
                  return (VIF_RET_INVALIDPARAMETER);
            case DMA_RET_WRONGALIGNMENT:
                  return (VIF_RET_WRONGALIGNMENT);
            case DMA_RET_DATATOOBIGTOFIT:
                  DEBUG_Print ("ERROR: VIF buffer overflow\n");
                  return (VIF_RET_DATATOOBIGTOFIT);
            default:
                  break;
         }

         // Add the data to the DMA chain (NOTE: Here I assume VIF_ATTACHENT_xxx
         // are the same as DMA_ATTACHMENT_xxx.  Bad me)
         switch (DMA_ChainAdd (VIF_DoubleBuffer[dest], buffer,
                               255, attachment))
         {
            case DMA_RET_INVALIDPARAMETER:
                  return (VIF_RET_INVALIDPARAMETER);
            case DMA_RET_WRONGALIGNMENT:
                  return (VIF_RET_WRONGALIGNMENT);
            case DMA_RET_DATATOOBIGTOFIT:
                  DEBUG_Print ("ERROR: VIF buffer overflow\n");
                  return (VIF_RET_DATATOOBIGTOFIT);
            default:
                  break; 
         }

         buffersizeinqw -= 255;
         buffer += 255;
         destvuaddr += 255;
      }
      else
      {
         viftag[3].num = buffersizeinqw;
         viftag[3].immediate = destvuaddr;

         // Add the tag to the DMA chain
         switch (DMA_ChainAdd (VIF_DoubleBuffer[dest], (OS_Qword *)viftag,
                               sizeof (viftag) / sizeof (OS_Qword), DMA_ATTACHMENT_INLINE))
         {
            case DMA_RET_INVALIDPARAMETER:
                  return (VIF_RET_INVALIDPARAMETER);
            case DMA_RET_WRONGALIGNMENT:
                  return (VIF_RET_WRONGALIGNMENT);
            case DMA_RET_DATATOOBIGTOFIT:
                  DEBUG_Print ("ERROR: VIF buffer overflow\n");
                  return (VIF_RET_DATATOOBIGTOFIT);
            default:
                  break;
         }

         // Add the data to the DMA chain (NOTE: Here I assume VIF_ATTACHENT_xxx
         // are the same as DMA_ATTACHMENT_xxx.  Bad me)
         switch (DMA_ChainAdd (VIF_DoubleBuffer[dest], buffer,
                               buffersizeinqw, attachment))
         {
            case DMA_RET_INVALIDPARAMETER:
                  return (VIF_RET_INVALIDPARAMETER);
            case DMA_RET_WRONGALIGNMENT:
                  return (VIF_RET_WRONGALIGNMENT);
            case DMA_RET_DATATOOBIGTOFIT:
                  DEBUG_Print ("ERROR: VIF buffer overflow\n");
                  return (VIF_RET_DATATOOBIGTOFIT);
            default:
                  break; 
         }

         buffersizeinqw = 0;
      }
   }

   // Everything was fine
   return (VIF_RET_OK);
}
                            
//--------------------------------------------------
// Add a tag to be sent to the VIF to the DMA chain.
// Note that if you're planning on adding more than
// 1 tag, use "VIF_AddTags" instead.  It will save
// some space.
//--------------------------------------------------
unsigned int VIF_AddTag (unsigned int command, unsigned int num,
                         unsigned int immediate, unsigned int dest)
{
   static VIF_TAG  viftag[4] = {{0, 0, VIF_CMD_NOP},
                                {0, 0, VIF_CMD_NOP},
                                {0, 0, VIF_CMD_NOP},
                                {0, 0, VIF_CMD_NOP}};

   if (VIF_Initialized == 0)
      return (VIF_RET_NOTINITIALIZED);

   // Make sure parameters are valid
   if ((command >= VIF_CMD_MAX) || (num >= 256) ||
       (immediate > 0xFFFF) || (dest >= VIF_DEST_MAX))
      return (VIF_RET_INVALIDPARAMETER);

   // Fill in the last of the 4 vif tags.  We can't just
   // send 1 tag to our DMAchain, as "DMA_ChainAdd"
   // requires buffersize to be a multiple of 16 bytes.
   viftag[3].cmd = command;
   viftag[3].num = num;
   viftag[3].immediate = immediate;

   // Add the tag to the DMA chain
   switch (DMA_ChainAdd (VIF_DoubleBuffer[dest], (OS_Qword *)viftag,
                         sizeof (viftag) / sizeof (OS_Qword), DMA_ATTACHMENT_INLINE))
   {
      case DMA_RET_INVALIDPARAMETER:
            return (VIF_RET_INVALIDPARAMETER);
      case DMA_RET_WRONGALIGNMENT:
            return (VIF_RET_WRONGALIGNMENT);
      case DMA_RET_DATATOOBIGTOFIT:
            DEBUG_Print ("ERROR: VIF buffer overflow\n");
            return (VIF_RET_DATATOOBIGTOFIT);
      default:
            break;
   }

   // Everything was fine
   return (VIF_RET_OK);
}

//---------------------------------------------------
// Add 4 tags to be sent to the VIF to the DMA chain.
//---------------------------------------------------
unsigned int VIF_Add4Tags (unsigned int command1, unsigned int num1,
                           unsigned int immediate1,
                           unsigned int command2, unsigned int num2,
                           unsigned int immediate2,
                           unsigned int command3, unsigned int num3,
                           unsigned int immediate3,
                           unsigned int command4, unsigned int num4,
                           unsigned int immediate4,
                           unsigned int dest)
{
   VIF_TAG  viftag[4];

   if (VIF_Initialized == 0)
      return (VIF_RET_NOTINITIALIZED);

   // Make sure parameters are valid
   if ((command1 >= VIF_CMD_MAX) || (num1 >= 256) ||
       (immediate1 > 0xFFFF) ||
       (command2 >= VIF_CMD_MAX) || (num2 >= 256) ||
       (immediate2 > 0xFFFF) ||
       (command3 >= VIF_CMD_MAX) || (num3 >= 256) ||
       (immediate3 > 0xFFFF) ||
       (command4 >= VIF_CMD_MAX) || (num4 >= 256) ||
       (immediate4 > 0xFFFF) || (dest >= VIF_DEST_MAX))
      return (VIF_RET_INVALIDPARAMETER);

   // Fill in the vif tags
   {
      viftag[0].cmd = command1;
      viftag[0].num = num1;
      viftag[0].immediate = immediate1;

      viftag[1].cmd = command2;
      viftag[1].num = num2;
      viftag[1].immediate = immediate2;

      viftag[2].cmd = command3;
      viftag[2].num = num3;
      viftag[2].immediate = immediate3;

      viftag[3].cmd = command4;
      viftag[3].num = num4;
      viftag[3].immediate = immediate4;
   }

   // Add the tag to the DMA chain
   switch (DMA_ChainAdd (VIF_DoubleBuffer[dest], (OS_Qword *)viftag,
                         sizeof (viftag) / sizeof (OS_Qword), DMA_ATTACHMENT_INLINE))
   {
      case DMA_RET_INVALIDPARAMETER:
            return (VIF_RET_INVALIDPARAMETER);
      case DMA_RET_WRONGALIGNMENT:
            return (VIF_RET_WRONGALIGNMENT);
      case DMA_RET_DATATOOBIGTOFIT:
            DEBUG_Print ("ERROR: VIF buffer overflow\n");
            return (VIF_RET_DATATOOBIGTOFIT);
      default:
            break;
   }

   // Everything was fine
   return (VIF_RET_OK);
}

//------------------------------------------------------------------------
// Add a texture of specified width, height and bitdepth to the DMA chain.
//------------------------------------------------------------------------
unsigned int VIF_AddTextureAndClut (OS_Qword *texture, OS_Qword *clut,
                                    unsigned int width, unsigned int height,
                                    unsigned int bitdepth, unsigned int vramaddresstexture,
                                    unsigned int vramaddressclut)
{
   OS_Qword      GSRegsSetup[8];
   unsigned int  dbw, clutdbw;
   unsigned int  texturepacketsize;
   unsigned int  clutpacketsize;
   unsigned int  clutwidth, clutheight;
   VIF_TAG       *tags;

   // At least a texture or a clut must be specified
   if ((texture == NULL) && (clut == NULL))
      return (VIF_RET_INVALIDPARAMETER);

   dbw = width >> 6;
   if (dbw == 0)
      dbw = 1;
   else if (dbw > 32)
      return (VIF_RET_INVALIDPARAMETER);

   switch (bitdepth)
   {
      case PS2_GS_PSMCT32:
            texturepacketsize = (width * height) >> 2;
            clutpacketsize = 0;
            break;

      case PS2_GS_PSMCT16:
            texturepacketsize = (width * height) >> 3;
            clutpacketsize = 0;
            break;

      case PS2_GS_PSMT8:
            texturepacketsize = (width * height) >> 4;
            clutpacketsize = 256 >> 2;
            clutwidth = 16;
            clutheight = 16;
            clutdbw = 1;
            break;

      case PS2_GS_PSMT4:
            texturepacketsize = (width * height) >> 5;
            clutpacketsize = 16 >> 2;
            clutwidth = 8;
            clutheight = 2;
            clutdbw = 1;
            break;

      default:
            return (VIF_RET_INVALIDPARAMETER);
   }

   // Only upload a clut if there's one...
   if ((clut != NULL) && (clutpacketsize))
   {
      // Let GS know it's about to receive a clut
      {
         // Need a DIRECT VIF tag
         tags = (VIF_TAG *)(&GSRegsSetup[0]);
         tags[0].cmd = VIF_CMD_NOP;
         tags[0].num = 0;
         tags[0].immediate = 0;
         tags[1].cmd = VIF_CMD_NOP;
         tags[1].num = 0;
         tags[1].immediate = 0;
         tags[2].cmd = VIF_CMD_NOP;
         tags[2].num = 0;
         tags[2].immediate = 0;
         tags[3].cmd = VIF_CMD_DIRECT;
         tags[3].num = 0;
         tags[3].immediate = 6;

         // GIFTag (for registers)
         GSRegsSetup[1].ui64[0] = PS2_GIFTAG_SET_TOPHALF (1, 0, 0, 0, PS2_GIFTAG_FLG_PACKED, 4);
         GSRegsSetup[1].ui64[1] = 0x000000000000EEEE;

         // BITBLTBUF
         GSRegsSetup[2].ui64[0] = PS2_GS_SETREG_BITBLTBUF (0, 0, 0, vramaddressclut, clutdbw, PS2_GS_PSMCT32);
         GSRegsSetup[2].ui64[1] = PS2_GS_BITBLTBUF;

         // TRXPOS
         GSRegsSetup[3].ui64[0] = PS2_GS_SETREG_TRXPOS (0, 0, 0, 0, 0);
         GSRegsSetup[3].ui64[1] = PS2_GS_TRXPOS;

         // TRXREG
         GSRegsSetup[4].ui64[0] = PS2_GS_SETREG_TRXREG (clutwidth, clutheight);
         GSRegsSetup[4].ui64[1] = PS2_GS_TRXREG;

         // TRXDIR
         GSRegsSetup[5].ui64[0] = PS2_GS_SETREG_TRXDIR (0);
         GSRegsSetup[5].ui64[1] = PS2_GS_TRXDIR;

         // GifTag (for texture image)
         GSRegsSetup[6].ui64[0] = PS2_GIFTAG_SET_TOPHALF ((clutwidth * clutheight) >> 2, texture ? 0 : 1, 0, 0, PS2_GIFTAG_FLG_IMAGE, 0);
         GSRegsSetup[6].ui64[1] = 0x0000000000000000;
      }

      // Need a DIRECT VIF tag
      tags = (VIF_TAG *)(&GSRegsSetup[7]);
      tags[0].cmd = VIF_CMD_NOP;
      tags[0].num = 0;
      tags[0].immediate = 0;
      tags[1].cmd = VIF_CMD_NOP;
      tags[1].num = 0;
      tags[1].immediate = 0;
      tags[2].cmd = VIF_CMD_NOP;
      tags[2].num = 0;
      tags[2].immediate = 0;
      tags[3].cmd = VIF_CMD_DIRECT;
      tags[3].num = 0;
      tags[3].immediate = (clutwidth * clutheight) >> 2;

      // Add the register settings in the DMA chain
      switch (DMA_ChainAdd (VIF_DoubleBuffer[1], GSRegsSetup, 8, DMA_ATTACHMENT_INLINE))
      {
         case DMA_RET_INVALIDPARAMETER:
               return (VIF_RET_INVALIDPARAMETER);
         case DMA_RET_WRONGALIGNMENT:
               return (VIF_RET_WRONGALIGNMENT);
         case DMA_RET_DATATOOBIGTOFIT:
               DEBUG_Print ("ERROR: VIF buffer overflow\n");
               return (VIF_RET_DATATOOBIGTOFIT);
         default:
               break;
      }

      // Add the register settings in the DMA chain
      switch (DMA_ChainAdd (VIF_DoubleBuffer[1], clut, (clutwidth * clutheight) >> 2, DMA_ATTACHMENT_INLINE/*DMA_ATTACHMENT_REFERENCE*/)) // If you can guarantee all cluts are compiled in an ELF, reference is better
      {
         case DMA_RET_INVALIDPARAMETER:
               return (VIF_RET_INVALIDPARAMETER);
         case DMA_RET_WRONGALIGNMENT:
               return (VIF_RET_WRONGALIGNMENT);
         case DMA_RET_DATATOOBIGTOFIT:
               DEBUG_Print ("ERROR: VIF buffer overflow\n");
               return (VIF_RET_DATATOOBIGTOFIT);
         default:
               break;
      }
   }

   if (texture == NULL)
   {
      // No need to go further, there's no texture
      return (VIF_RET_OK);
   }

   // Need a DIRECT VIF tag
   tags = (VIF_TAG *)(&GSRegsSetup[0]);
   tags[0].cmd = VIF_CMD_NOP;
   tags[0].num = 0;
   tags[0].immediate = 0;
   tags[1].cmd = VIF_CMD_NOP;
   tags[1].num = 0;
   tags[1].immediate = 0;
   tags[2].cmd = VIF_CMD_NOP;
   tags[2].num = 0;
   tags[2].immediate = 0;
   tags[3].cmd = VIF_CMD_DIRECT;
   tags[3].num = 0;
   tags[3].immediate = 6;

   // GIFTag (for registers)
   GSRegsSetup[1].ui64[0] = PS2_GIFTAG_SET_TOPHALF (1, 0, 0, 0, PS2_GIFTAG_FLG_PACKED, 4);
   GSRegsSetup[1].ui64[1] = 0x000000000000EEEE;

   // BITBLTBUF
   GSRegsSetup[2].ui64[0] = PS2_GS_SETREG_BITBLTBUF (0, 0, 0, vramaddresstexture, dbw, bitdepth);
   GSRegsSetup[2].ui64[1] = PS2_GS_BITBLTBUF;

   // TRXPOS
   GSRegsSetup[3].ui64[0] = PS2_GS_SETREG_TRXPOS (0, 0, 0, 0, 0);
   GSRegsSetup[3].ui64[1] = PS2_GS_TRXPOS;

   // TRXREG
   GSRegsSetup[4].ui64[0] = PS2_GS_SETREG_TRXREG (width, height);
   GSRegsSetup[4].ui64[1] = PS2_GS_TRXREG;

   // TRXDIR
   GSRegsSetup[5].ui64[0] = PS2_GS_SETREG_TRXDIR (0);
   GSRegsSetup[5].ui64[1] = PS2_GS_TRXDIR;

   // Can we upload the texture in one chunk?
   if (texturepacketsize < 32768)
   {
      // The texture is small enough to be uploaded in one chunk

      // GifTag (for texture image)
      GSRegsSetup[6].ui64[0] = PS2_GIFTAG_SET_TOPHALF (texturepacketsize, 1, 0, 0, PS2_GIFTAG_FLG_IMAGE, 0);
      GSRegsSetup[6].ui64[1] = 0x0000000000000000;

      // Need a DIRECT VIF tag
      tags = (VIF_TAG *)(&GSRegsSetup[7]);
      tags[0].cmd = VIF_CMD_NOP;
      tags[0].num = 0;
      tags[0].immediate = 0;
      tags[1].cmd = VIF_CMD_NOP;
      tags[1].num = 0;
      tags[1].immediate = 0;
      tags[2].cmd = VIF_CMD_NOP;
      tags[2].num = 0;
      tags[2].immediate = 0;
      tags[3].cmd = VIF_CMD_DIRECT;
      tags[3].num = 0;
      tags[3].immediate = texturepacketsize;

      // Add the register settings in the DMA chain
      switch (DMA_ChainAdd (VIF_DoubleBuffer[1], GSRegsSetup, 8, DMA_ATTACHMENT_INLINE))
      {
         case DMA_RET_INVALIDPARAMETER:
               return (VIF_RET_INVALIDPARAMETER);
         case DMA_RET_WRONGALIGNMENT:
               return (VIF_RET_WRONGALIGNMENT);
         case DMA_RET_DATATOOBIGTOFIT:
               DEBUG_Print ("ERROR: VIF buffer overflow\n");
               return (VIF_RET_DATATOOBIGTOFIT);
         default:
               break;
      }

      // Add a reference to the texture in the DMA chain
      switch (DMA_ChainAdd (VIF_DoubleBuffer[1], texture, texturepacketsize, DMA_ATTACHMENT_INLINE/*DMA_ATTACHMENT_REFERENCE*/)) // If you can guarantee all textures are compiled in an ELF, reference is better
      {
         case DMA_RET_INVALIDPARAMETER:
               return (VIF_RET_INVALIDPARAMETER);
         case DMA_RET_WRONGALIGNMENT:
               return (VIF_RET_WRONGALIGNMENT);
         case DMA_RET_DATATOOBIGTOFIT:
               DEBUG_Print ("ERROR: VIF buffer overflow\n");
               return (VIF_RET_DATATOOBIGTOFIT);
         default:
               break;
      }
   }
   else
   {
      // The GIF has a limit on the packet size it can send (32767 qwords
      // to be exact). So big textures have to be sent in smaller chunks...

      // We'll only transfer 5 qwords after all...
      tags[3].immediate = 5;

      // Add the register settings in the DMA chain
      switch (DMA_ChainAdd (VIF_DoubleBuffer[1], GSRegsSetup, 6, DMA_ATTACHMENT_INLINE))
      {
         case DMA_RET_INVALIDPARAMETER:
               return (VIF_RET_INVALIDPARAMETER);
         case DMA_RET_WRONGALIGNMENT:
               return (VIF_RET_WRONGALIGNMENT);
         case DMA_RET_DATATOOBIGTOFIT:
               DEBUG_Print ("ERROR: VIF buffer overflow\n");
               return (VIF_RET_DATATOOBIGTOFIT);
         default:
               break;
      }

      while (texturepacketsize)
      {
         if (texturepacketsize <= 16384)
         {
            // Need a DIRECT VIF tag
            tags = (VIF_TAG *)(&GSRegsSetup[0]);
            tags[0].cmd = VIF_CMD_NOP;
            tags[0].num = 0;
            tags[0].immediate = 0;
            tags[1].cmd = VIF_CMD_NOP;
            tags[1].num = 0;
            tags[1].immediate = 0;
            tags[2].cmd = VIF_CMD_NOP;
            tags[2].num = 0;
            tags[2].immediate = 0;
            tags[3].cmd = VIF_CMD_DIRECT;
            tags[3].num = 0;
            tags[3].immediate = texturepacketsize + 1;

            // GIFTag (for registers)
            GSRegsSetup[1].ui64[0] = PS2_GIFTAG_SET_TOPHALF (texturepacketsize, 0, 0, 0, PS2_GIFTAG_FLG_IMAGE, 0);
            GSRegsSetup[1].ui64[1] = 0x0000000000000000;

            texturepacketsize = 0;
         }
         else
         {
            // Need a DIRECT VIF tag
            tags = (VIF_TAG *)(&GSRegsSetup[0]);
            tags[0].cmd = VIF_CMD_NOP;
            tags[0].num = 0;
            tags[0].immediate = 0;
            tags[1].cmd = VIF_CMD_NOP;
            tags[1].num = 0;
            tags[1].immediate = 0;
            tags[2].cmd = VIF_CMD_NOP;
            tags[2].num = 0;
            tags[2].immediate = 0;
            tags[3].cmd = VIF_CMD_DIRECT;
            tags[3].num = 0;
            tags[3].immediate = 16384 + 1;

            // GIFTag (for registers)
            GSRegsSetup[1].ui64[0] = PS2_GIFTAG_SET_TOPHALF (16384, 0, 0, 0, PS2_GIFTAG_FLG_IMAGE, 0);
            GSRegsSetup[1].ui64[1] = 0x0000000000000000;

            texturepacketsize -= 16384;
         }

         // Add the register settings in the DMA chain
         switch (DMA_ChainAdd (VIF_DoubleBuffer[1], GSRegsSetup, 2, DMA_ATTACHMENT_INLINE))
         {
            case DMA_RET_INVALIDPARAMETER:
                  return (VIF_RET_INVALIDPARAMETER);
            case DMA_RET_WRONGALIGNMENT:
                  return (VIF_RET_WRONGALIGNMENT);
            case DMA_RET_DATATOOBIGTOFIT:
                  DEBUG_Print ("ERROR: VIF buffer overflow\n");
                  return (VIF_RET_DATATOOBIGTOFIT);
            default:
                  break;
         }

         // Add a reference to the texture in the DMA chain
         switch (DMA_ChainAdd (VIF_DoubleBuffer[1], texture, GSRegsSetup[1].ui64[0] & 0x7FFF, DMA_ATTACHMENT_INLINE/*DMA_ATTACHMENT_REFERENCE*/))  // If you can guarantee all textures are compile in an ELF, reference is better
         {
            case DMA_RET_INVALIDPARAMETER:
                  return (VIF_RET_INVALIDPARAMETER);
            case DMA_RET_WRONGALIGNMENT:
                  return (VIF_RET_WRONGALIGNMENT);
            case DMA_RET_DATATOOBIGTOFIT:
                  DEBUG_Print ("ERROR: VIF buffer overflow\n");
                  return (VIF_RET_DATATOOBIGTOFIT);
            default:
                  break;
         }

         texture += 4096;  // texture is an unsigned int pointer, not OS_Qword pointer...
      }
   }

   // Everything was fine
   return (VIF_RET_OK);
}

//-------------------------------------
// Initiate the DMA chain transfer,
//-------------------------------------
unsigned int VIF_SwapBuffers (unsigned int dest)
{
   if (VIF_Initialized == 0)
      return (VIF_RET_NOTINITIALIZED);

   // Make sure the parameter is valid
   if (dest >= VIF_DEST_MAX)
      return (VIF_RET_INVALIDPARAMETER);

   // Initiate the DMA chain transfer
   DMA_ChainSend (VIF_DoubleBuffer[dest]);

   // Everything was fine
   return (VIF_RET_OK);
}

//------------------------------------------
// Wait for end of transfer of the VIF chain
//------------------------------------------
unsigned int VIF_TransferWait (unsigned int dest)
{
   if (VIF_Initialized == 0)
      return (VIF_RET_NOTINITIALIZED);

   // Make sure the parameter is valid
   if (dest >= VIF_DEST_MAX)
      return (VIF_RET_INVALIDPARAMETER);

   // Initiate the DMA chain transfer
   DMA_ChainTransferWait (VIF_DoubleBuffer[dest]);

   // Everything was fine
   return (VIF_RET_OK);
}

//------------------------------------------------
// Return the current size of the chain, in qwords
//------------------------------------------------
unsigned int VIF_GetChainSize (unsigned int dest, unsigned int *chainsize)
{
   if (VIF_Initialized == 0)
      return (VIF_RET_NOTINITIALIZED);

   // Make sure parameters are valid
   if ((dest >= VIF_DEST_MAX) || (chainsize == NULL))
      return (VIF_RET_INVALIDPARAMETER);

   if (DMA_GetChainSize (VIF_DoubleBuffer[dest], chainsize) != DMA_RET_OK)
      return (VIF_RET_ERROR);

   // Everything was fine
   return (VIF_RET_OK);
}
